"""Small module responsible for checking whether standard
Grid tools exist and whether the GRID credentials are
activated
"""
import subprocess
from functools import wraps


def check_uberftp():
    """Checks if the gfal-ls executable
    exists on the system. Returns True if it exists

    :returns: bool"""
    process = subprocess.Popen(['which', 'gfal-ls'], stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            encoding='utf8')
    output = process.communicate()
    if output[0] == '' and output[1] == '':
        return False
    return True

def grid_credentials_enabled():
    """Short function that checks whether
    the GRID credentials have been enabled or expired
    Returns True if they are currenctly active

    This requires gfal-ls!"""
    if not check_uberftp():
        return False
    process = subprocess.Popen([
        'gfal-ls', '-l',
        'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/diskonly'
        ], stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        encoding='utf8')
    res = process.communicate()
    if type(res[1])==bytes:
        error = res[1].decode('utf8')
    else:
        error = res[1]
    if "Failed to acquire credentials." in error:
        raise Exception("Grid Credentials expired! "
                        "Run 'startGridSession lofar:/lofar/user/sksp' in the shell")
    return True


def skip_grid_auth(func):
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except :
            raise Warning("Not authorized GRID user but continuing anyways")
    return wrapped




import sys
import re
from collections import deque
from math import ceil
import logging
import subprocess

GSI_FQDNs={
        'grid.sara.nl':'sara',
        'fz-juelich.de':'juelich',
        'lofar.psnc.pl':'poznan'
        }

class srmlist(list):
    """
    The srmlist class is an extension of Python lists that can hold a list of
    srm links to data on GRID Storage (LOFAR Archive, Intermediate Storage, etc).

    In addition to the regular list capabilities, it also has internal checks for the
    location and the OBSID of the data. When a new item is appended, these checks are
    done automatically. Checking OBSID is an optional argument set to True by default.
    """

    def __init__(self, check_OBSID=True, check_location=True, link=None):
        """__init__: Initializes the srmlist object. 

        :param check_OBSID: Boolean flag to check if each added link has the same OBSID
        :type check_OBSID: Boolean
        :param check_location: Boolean flag to check if all files are in the same location (for staging purposes)
        :type check_location: Boolean
        :param link: append a link to the srmlist at creation
        :type link: str
        """
        super(srmlist, self).__init__()
        self._check_location = check_location
        self.lta_location = None
        self.obsid = None
        self.checkobsid = check_OBSID
        if link:
            self.append(link)

    def check_link_location(self, item):
        """Checks the location of the item"""
        tmp_loc = ""
        if isinstance(item, str):
            tmp_loc = self.check_str_location(item)
        elif isinstance(item, srmlist):
            for i in item:
                tmp_loc = self.check_str_location(i)
        return tmp_loc
    
    def from_file(self, filename):
        """You can automatically load a file into the srmlist using
        s = srmlist().from_file(filename)"""
        for line in open(filename,'r').readlines():
            self.append(line.strip('\n'))
        return self

    @property
    def LTA_location(self):
        if len(self)>0:
            return self.check_str_location(self[0])
    
    def check_str_location(self, item):
        """searches the item for an FQDN and returns the location
        of the data using the list below. Returns the data location or None"""
        for fqdn in GSI_FQDNs:
            if fqdn in item:
                return GSI_FQDNs[fqdn] 

    def stringify_item(self, item):
        if isinstance(item, str):
            link = item.strip('\n')
            link = item.strip('\r')
        elif isinstance(item, srmlist):
            link = "".join(str(v) for v in item)
        else:
            return ""
        return link

    def _check_obsid(self, item):
        link = self.stringify_item(item)
        tmp_obsid = re.search('L[0-9][0-9]*',
                              link).group(0)
        if not self.obsid:
            self.obsid = tmp_obsid
        if self.checkobsid and tmp_obsid != self.obsid:
            raise AttributeError("Different OBSID than previous items")

    def append(self, item):
        if not item or item == "":
            return
        if self.checkobsid:
            self._check_obsid(item)
        tmp_loc = self.check_link_location(item)
        item = self.trim_spaces(self.stringify_item(item))
        if not self.lta_location:
            self.lta_location = tmp_loc
        elif self.lta_location != tmp_loc  and self._check_location :
            raise AttributeError(
                "Appended srm link not the same location as previous links!")
        if item in self:
            return  # don't add duplicate srms
        # append the item to itself (the list)
        super(srmlist, self).append(item)

    def trim_spaces(self, item):
        """Sometimes there are two fields in the incoming list. Only take the first
        as long as it's fromatted properly
        """
        item = re.sub('//pnfs', '/pnfs', "".join(item))
        if self.lta_location == 'poznan':
            item = re.sub('//lofar', '/lofar', "".join(item))
        if " " in item:
            for potential_link in item.split(" "):
                if 'srm://' in potential_link:
                    return potential_link
        else:
            return item

    def gfal_replace(self, item):
        """
        For each item, it creates a valid link for the gfal staging scripts
        """
        if 'srm://' in item:
            return re.sub(':8443', ':8443/srm/managerv2?SFN=', item)
        elif 'gsiftp://' in item:
            return self.srm_replace(item)

    def srm_replace(self, item):
        if self.lta_location == 'sara':
            return re.sub('gsiftp://gridftp.grid.sara.nl:2811',
                          'srm://srm.grid.sara.nl:8443',
                          item)
        if self.lta_location == 'juelich':
            return re.sub("gsiftp://lofar-gridftp.fz-juelich.de:2811",
                          "srm://lofar-srm.fz-juelich.de:8443",
                          item)
        if self.lta_location == 'poznan':
            return re.sub("gsiftp://gridftp.lofar.psnc.pl:2811",
                          "srm://lta-head.lofar.psnc.pl:8443",
                          item)

    def gsi_replace(self, item):
        if self.lta_location == 'sara':
            return re.sub('srm://srm.grid.sara.nl:8443',
                          'gsiftp://gridftp.grid.sara.nl:2811',
                          item)
        if self.lta_location == 'juelich':
            return re.sub("srm://lofar-srm.fz-juelich.de:8443",
                    "gsiftp://lofar-gridftp.fz-juelich.de:2811", item)
        if self.lta_location == 'poznan':
            return re.sub("srm://lta-head.lofar.psnc.pl:8443",
                          "gsiftp://gridftp.lofar.psnc.pl:2811",
                          item)

    def http_replace(self, item):
        if self.lta_location == 'sara':
            return re.sub('srm://',
                          'https://lofar-download.grid.sara.nl/lofigrid/SRMFifoGet.py?surl=srm://',
                          item)
        if self.lta_location == 'juelich':
            return re.sub(
                "srm://",
                "https://lofar-download.fz-juelich.de/webserver-lofar/SRMFifoGet.py?surl=srm://",
                item)
        if self.lta_location == 'poznan':
            return re.sub("srm://",
                          "https://lta-download.lofar.psnc.pl/lofigrid/SRMFifoGet.py?surl=srm://",
                          item)

    def gsi_links(self):
        """
        Returns a generator which can be iterated over, this generator will return
        a set of gsiftp:// links which can be used with globus-url-copy and uberftp
        """
        queue = deque(self)
        while queue:
            item = queue.pop()
            if item:
                yield self.gsi_replace(item)

    def http_links(self):
        """
        Returns a generator that can be used to generate http:// links that can be downloaded
        using wget
        """
        queue = deque(self)
        while queue:
            item = queue.pop()
            if item:
                yield self.http_replace(item)

    def gfal_links(self):
        """
        Returns a generator that can be used to generate links that can be staged/stated with gfal
        """
        queue = deque(self)
        while queue:
            item = queue.pop()
            if item:
                yield self.gfal_replace(item)

    def sbn_dict(self, pref="SB", suff="_"):
        """
        Returns a generator that creates a pair of SBN and link. Can be used to create dictionaries
        """
        for i in self:
            match = None
            surl = srmlist()
            surl.append(i)
            match = re.search(pref+'(.+?)'+suff, i)
            try:
                yield match.group(1), surl
            except AttributeError as exc:
                sys.stderr.write("Are you using pref='SB' and suff='_'"+
                                 "to match ...SB000_... ?")
                raise exc


def slice_dicts(srmdict, slice_size=10):
    """
    Returns a dict of lists that hold 10 SBNs (by default).
    Missing Subbands are treated as empty spaces, if you miss SB009,
    the list will include  9 items from SB000 to SB008, and next will start at SB010"""
    srmdict = dict(srmdict)

    keys = sorted(srmdict.keys())
    start = int(keys[0])
    sliced = {}
    for chunk in range(0, 1 + int(ceil((int(keys[-1])-int(keys[0]))/float(slice_size)))):
        chunk_name = format(start+chunk*slice_size, '03')
        sliced[chunk_name] = srmlist()
        for i in range(slice_size):
            if format(start+chunk*slice_size+i, '03') in srmdict.keys():
                sliced[chunk_name].append(
                    srmdict[format(start+chunk*slice_size+i, '03')])
    sliced = dict((k, v) for k, v in sliced.items() if v) #Removing empty items
    return sliced

def make_srmlist_from_gsiftpdir(gsiftpdir):
    from GRID_LRT.storage import gsifile
    srml = srmlist()
    grid_dir = gsifile.GSIFile(gsiftpdir)
    for i in [f.loc for f in grid_dir.list_dir()]:
        srml.append(i)
    return srml

def count_files_uberftp(directory):
    from GRID_LRT.storage import gsifile
    grid_dir = gsifile.GSIFile(directory)
    return [f.location for f in grid_dir.list_dir()]



from datetime import datetime
from datetime import timedelta
from time import strptime
import warnings
import sys
import subprocess
from subprocess import Popen
import re
import GRID_LRT.auth.grid_credentials as grid_creds
from GRID_LRT.auth.get_picas_credentials import picas_cred
from GRID_LRT.Staging.srmlist import srmlist
from GRID_LRT.storage.utils import SafePopen
import humanfriendly

class SafePopen(Popen):
    def __init__(self, *args, **kwargs):
        if sys.version_info.major == 3 :
            kwargs['encoding'] = 'utf8'
        return super(SafePopen, self).__init__(*args, **kwargs)

import pdb
class GSIFile(object):
    def __init__(self, location, parent_dir=None):
        _ = grid_creds.grid_credentials_enabled()
        self.location = location
        self.protocol = location.split("://")[0]
        self.port = self._get_port(location)

        if parent_dir:
            self._build_from_parent_dir(parent_dir)
        else:
            self._subfiles = None
            self.is_dir, self.parent_dir = self._check_if_directory(location)
            self._internal = self._test_file_exists(location)
        self.datetime = self._extract_date(self._internal)
        self.size = self._get_size()
        if self.is_dir:
            self.is_file = False
            self.filename = self.location.split('/')[-1]
        else:
            self.is_file = True
            self.filename = self._internal[-1]

    def _build_from_parent_dir(self, parent_dir):
        self.parent_dir = parent_dir.location
        self._internal = self._find_item_in_uberftp_result(parent_dir._subfiles)
        if self._internal[0][0] == 'd':
            self.is_dir = True
        elif self._internal[0][0] == '-':
            self.is_dir = False

    def _get_size(self):
        self._bytesize = self._internal[4]
        human_size = humanfriendly.parse_size(self._bytesize)
        return humanfriendly.format_size(human_size)

    def get_dir_size(self):
        """Will go one level deeper and sum the sizes of all objects. 
    If the objects are directories, you'll have to do the iteration yourself!
        """
        total = humanfriendly.parse_size(self.size)
        for f in self.list_dir():
            total += humanfriendly.parse_size(f.size)
        print(total)
        t = humanfriendly.parse_size(str(total))
        return  humanfriendly.format_size(t)

    def _check_if_directory(self, location):
        self.num_subdir = len([i for i in location.split('gsiftp://')[1].split('/') if i]) - 1
        filename = location.split('/')[-1]
        parent_dir = self.get_parent_dir()
        parent_dir = parent_dir.replace(self.protocol+':/',self.protocol+"://")
        #TODO Make this cleaner (new staticmethod)
        res, err = self._uberftpls(parent_dir)
        if '550' in err:
            warnings.warn('Parent directory inaccessible!: {}'.format(err))
            return None, parent_dir #Doesn't know if it's a file or dir if can't get parent_dir
        if type(res) == bytes:
            res = res.decode('utf8')
        status = [x for x in res.split('\r\n') if filename in x]
        if status[0][0] == 'd':
            return True, parent_dir
        else:
            return False, parent_dir

    def __repr__(self):
        return "<GRID_LRT.storage.gsifile.GSIFile {} {} located in {} >".format(
                "File" if self.is_file else "Folder",self.filename, self.location)

    def _test_file_exists(self, location):
        result, error = self._uberftpls(location)
        if result == "" and error == "":
            if location[-1] == '/':
                location = location[:-1]
                result, error = self._uberftpls("/".join(location.split('/')[:-1]))
        if "No match " in error:
            raise Exception("file %s cannot be found: %s"%(location, error))
        if self.is_dir:
            result, err = self._uberftpls(self.parent_dir)
            result = self._find_item_in_uberftp_result(result)
            return result
        return result.split()


    def _find_item_in_uberftp_result(self, result):
        if self.location[-1] == '/':
            location = self.location[:-1]
        else:
            location = self.location
        if type(result) == bytes:
            result = result.decode('ascii')
        for i in result.split('\r\n'):
            if location.split('/')[-1] in i:
                result = i
                break
        result = result.split()
        return result

    @staticmethod
    def _extract_date(data):
        if isinstance(data[-2], bytes):
            year = str(data[-2].decode('ascii'))
            month = data[-4].decode('ascii')
            day = data[-3].decode('ascii')
        else:
            year = str(data[-2])
            month = data[-4]
            day = data[-3]
        if year not in ['2021','2020','2019', '2018', '2017', '2016', '2015']:
            ## In this case the year is not specified.
            this_year = datetime.now().year
            if strptime(month,'%b').tm_mon > datetime.now().month:
                year = this_year -1
            else:
                year = this_year
            date = month + " " + day + " " + str(year)
            if isinstance(data[-2], bytes):
                time = data[-2].decode('ascii')
            else:
                time = data[-2]
        else:
            date = month + " " + day + " " + year
            time = "00:00"
        
        file_datetime = datetime.strptime(date+"-"+time, "%b %d %Y-%H:%M")
        return file_datetime

    @staticmethod
    def _get_port(location):
        try:
            port = re.search(':[0-9]{4}/',location).group(0)[1:][:-1]
            return port
        except AttributeError:
            return None

    def _donotdelete(self,location):
        """Raises Exception if you try to delete files or folders whose parent is
        one of these folders"""
        locations = ['gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/diskonly'
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/archive',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/diskonly',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/sksp_natalie',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/distrib',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/diskonly/sandbox',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/diskonly/pipelines',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/diskonly/pipelines/PiLL',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/diskonly/pipelines/SKSP',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/diskonly/pipelines/test',
                     'gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/diskonly/pipelines/NDPPP'
#                     'gsiftp://gridftp.grid.sara.n:2811l/pnfs/grid.sara.nl/data/lofar/user/sksp/pipelines/DISC'
                     ]
        if self.parent_dir in locations:
            raise Exception("FORBIDDEN to delete this {} because its parent is {}".format(
                            'file' if self.is_file else 'folder', self.parent_dir))

    def delete(self):
        parent_dir = self.get_parent_dir()
        self._donotdelete(parent_dir)
        if self.is_dir and not self.is_empty():
            raise Exception("Not allowed to delete a folder that isn't empty yet" )
        del_proc = SafePopen(['uberftp', '-rm', self.location],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        res, err = del_proc.communicate()
        if not err:
            return
        else: 
            warnings.warn('Deleting failed: {}'.format(err))

    def copy(self, other_location):
        if self.is_dir:
            raise Exception("We cannot copy an entire folder, loop over its contents instead")
        copy = SafePopen(['globus-url-copy',self.location, other_location.location+"/"], 
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE) 
        res, err = copy.communicate()
        if not err:
            return
        else:
            warnings.warn('Copying {0} failed: {1}'.format(self.filename, err))


    def get_parent_dir(self):
        location = "/".join([i for i in self.location.split('/') if i])
        parent_dir = "/".join(location.split('/')[:-1])
        return parent_dir

    def _get_num_files(self):
        """If self is folder, gets the number of files in it"""
        if self.is_file:
            return 1
        res, err = self._uberftpls(self.location)
        num_files = len([x for x in res.split('\r\n') if x])
        return num_files

    def _uberftpls(self, location):
        sub = SafePopen(['gfal-ls','-l', location],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        res, err = sub.communicate()
        if err:
            warnings.warn("gfal-ls gave us an error for location {0}: {1}".format(location, err))
        return res, err

    def is_empty(self):
        if self._get_num_files() == 0:
            return True
        return False

    def list_dir(self):
        """Checks the status of all srm links in the folder given. 
        Returns a list of localities
        """     
        if self.is_dir:
           results, error= self._uberftpls(self.location)
        else: 
            return []
        if results == '':
            return []
        results = results.strip().split("\n")
        file_locs = [self.location +"/"+str(i.split()[-1])
                for i in results if i]
        self._subfiles, _ = self._uberftpls(self.location)
        files_list = [GSIFile(i, parent_dir=self) for i, r in zip(file_locs, results)]
        return files_list


def gsi_mkdir(location):
    mkdir = SafePopen(['gfal-mkdir',location],
            stdout=subprocess.PIPE, stderr=subprocess.PIPE) 
    out, err = mkdir.communicate()
    if err:
        warnings.warn('mkdir {0} failed: {1}'.format(location, err))
    return GSIFile(location)



class MockGSIFile(GSIFile):
    def __init__(self, location):
        """uses text from uberftp -ls to initialize, saved in a file"""
        d = open(location, 'r').read()
        self._filedata = '\r\n'.join(d.split('\r\n')[1:])
        self.location = d.split('\r\n')[0]
        self.parent_dir = self.location.split('/')[-1]
        self._internal = self._find_item_in_uberftp_result(self._filedata)
        if len(d)>2:
            self.is_dir = True
            self.is_file = False
        else:
            self.is_dir = False
            self.is_file = True

    def list_dir(self):
        results = self._filedata.split('\r\n')[1:]
        file_locs = [self.location +"/"+str(i.split()[-1])
                                for i in results if i]
        return file_locs





"""
It uses an xmlrpc proxy to talk and authenticate to the remote service.
Your account credentials will be read from the awlofar catalog Environment.cfg,
if present or can be provided in a .stagingrc file in your home directory.

!!Please do not talk directly to the xmlrpc interface,
but use this module to access the provided functionality.
!! This is to ensure that when we change the remote interface,
your scripts don't break and you will only have to upgrade this module.
"""
from __future__ import print_function
import os
from os.path import expanduser
import datetime
from functools import wraps

try:
    import xmlrpclib
except ImportError:
    import xmlrpc.client as xmlrpclib  # pylint: disable=import-error

__version__ = "1.0"

# ---
# Determine credentials and create proxy
USER = None
PASSW = None

def get_staging_creds(user=None, passw=None):
    """Function to get the staging credentials first from
       ~/.awe/Environment.cfg, if not there, then from ~/.stagingrc
       and finally from the env variables LOFAR_LTA_USER and LOFAR_LTA_PWD
    """
    if user and passw:
        lta_proxy = xmlrpclib.ServerProxy(
                "https://"+user+':'+passw+"@webportal.astron.nl/service-public/xmlrpc")
        return user, passw, lta_proxy
    try:
        with open(expanduser("~/.awe/Environment.cfg"), 'r') as authfile:
            print(datetime.datetime.now(), "stager_access: Parsing user credentials from",
                  expanduser("~/.awe/Environment.cfg"))
            for line in authfile:
                if line.startswith("database_user"):
                    user = line.split(':')[1].strip()
                if line.startswith("database_password"):
                    passw = line.split(':')[1].strip()
    except IOError:
        try:
            with open(expanduser("~/.stagingrc"), 'r') as authfile:
                print(datetime.datetime.now(
                ), "stager_access: Parsing user credentials from", expanduser("~/.stagingrc"))
                for line in authfile:
                    if line.startswith("user"):
                        user = line.split('=')[1].strip()
                    if line.startswith("password"):
                        passw = line.split('=')[1].strip()
        except IOError:
            print("No StagingRC file found")
            try:
                user = os.environ['LOFAR_LTA_USER']
                passw = os.environ['LOFAR_LTA_PWD']
            except KeyError:
                print("LOFAR LTA USER/PASSW not in environment!")
    if user and passw:
        print(datetime.datetime.now(), "stager_access: Creating proxy")
        lta_proxy = xmlrpclib.ServerProxy(
            "https://"+user+':'+passw+"@webportal.astron.nl/service-public/xmlrpc")
    else:
        print("No User or Password exist. ")
        return "", "", ""
    return user, passw, lta_proxy

USER, PASSW, LTA_PROXY = get_staging_creds()
# ---


def handle_xmlrpc_exception(fun):
    """ Exception handler that stops xmlrpclib.ProtocolError
    from printing out the username and password to the terminal

    """
    @wraps(fun)
    def wrapper(*args, **kwds):
        """Wrapper around the function that captures the exception
        and DOES NOT print out the password in plain text"""
        try:
            return fun(*args, **kwds)
        except xmlrpclib.ProtocolError as err:
            if PASSW in err.url:
                err.url = err.url.replace(PASSW, '[REDACTED]')
            raise err
    return wrapper


@handle_xmlrpc_exception
def stage(surls):
    """ Stage list of SURLs or a string holding a single SURL

    :param surls: Either a list of strings or a string holding a single surl to stage
    :type surls: either a list() or a str()
    :return: An integer which is used to refer to the stagig request when polling
    the API for a staging status
    """
    staged_surls = []
    if isinstance(surls, str):
        staged_surls = [surls]
    for i in surls:
        staged_surls.append(str(i))
    stageid = LTA_PROXY.LtaStager.add_getid(staged_surls)
    return stageid


@handle_xmlrpc_exception
def get_status(stageid):
    """ Get status of request with given ID

    Args:
        :param stageid: The id of the staging request which you want the status of
        :type stageid: int

    Returns:
        :status: A string describing the staging status: 'new', 'scheduled',
        'in progress' or 'success'
        """
    return LTA_PROXY.LtaStager.getstatus(stageid)


@handle_xmlrpc_exception
def abort(stageid):
    """ Abort running request / release data of a finished request with given ID """
    return LTA_PROXY.LtaStager.abort(stageid)


@handle_xmlrpc_exception
def get_surls_online(stageid):
    """ Get a list of all files that are already online for a running request with given ID  """
    return LTA_PROXY.LtaStager.getstagedurls(stageid)


@handle_xmlrpc_exception
def get_srm_token(stageid):
    """ Get the SRM request token for direct interaction with the SRM site via Grid/SRM tools """
    return LTA_PROXY.LtaStager.gettoken(stageid)


@handle_xmlrpc_exception
def reschedule(stageid):
    """ Reschedule a request with a given ID, e.g. after it was put on hold due to maintenance """
    return LTA_PROXY.LtaStager.reschedule(stageid)


@handle_xmlrpc_exception
def get_progress():
    """ Get a detailed list of all running requests and their current progress.
    As a normal user, this only returns your own requests.  """
    return LTA_PROXY.LtaStager.getprogress()


@handle_xmlrpc_exception
def get_storage_info():
    """ Get storage information of the different LTA sites,
    e.g. to check available disk pool space. Requires support role permissions. """
    return LTA_PROXY.LtaStager.getsrmstorageinfo()


def prettyprint(dictionary, indent=""):
    """ Prints nested dict responses nicely. Example:
    'stager_access.prettyprint(stager_access.get_progress())'"""
    if isinstance(dictionary, dict):
        for key in sorted(dictionary.keys()):
            item = dictionary.get(key)
            if isinstance(item, dict):
                print(indent+'+ ' + str(key))
                prettyprint(item, indent=indent+'  ')
            else:
                print(indent+'- '+str(key), '\t -> \t', str(item))
    else:
        print("stager_access: This prettyprint takes a dict only!")



"""GRID_LRT: Grid LOFAR Tools"""
import sys
import os
import socket
from subprocess import call, Popen, PIPE, STDOUT
if sys.version_info[0:2] != (2, 6):
    from subprocess import check_output


__all__ = ["storage", 'auth', "application", "Staging", 'token']
__version__ = "1.0.5"
__author__ = "Alexandar P. Mechev"
__copyright__ = "2019 Alexandar P. Mechev"
__credits__ = ["Alexandar P. Mechev", "Natalie Danezi", "J.B.R. Oonk"]
__bibtex__ = """@misc{apmechev:2019,
      author       = {Alexandar P. Mechev} 
      title        = {apmechev/GRID_LRT: v1.0.0},
      month        = sep,
      year         = 2019,
      doi          = {10.5281/zenodo.1438833},
      url          = {https://doi.org/10.5281/zenodo.1438833}
    }"""
__license__ = "GPL 3.0"
__maintainer__ = "Alexandar P. Mechev"
__email__ = "apmechev+LOFAR@gmail.com"
__status__ = "Production"
__date__ = "2019-10-22"





import datetime
import GRID_LRT

def modify_updated_date(date=datetime.datetime.now().date()):
    """use the date file to save the latest update date"""
    mod_file = GRID_LRT.__file__.replace('pyc', 'py')
    with open(mod_file) as _file:
        file_data = _file.read()
    file_data = file_data.split('\n')
    for i in file_data:
        if '__date__' in i:
            idx = file_data.index(i)
            file_data[idx] = "__date__ = \"%s\"" %(date.isoformat())
    with open(mod_file, 'w') as _file:
        _file.write('\n'.join(file_data))

if __name__ == '__main__':
    modify_updated_date()



#!/usr/bin/env python

""" Staging script using the gfal API

# ===================================================================== #
# author: Ron Trompert <ron.trompert@surfsara.nl>	--  SURFsara    #
# helpdesk: Grid Services <grid.support@surfsara.nl>    --  SURFsara    #
#                                                                       #
# usage: python stage_all.py                                                #
# description:                                                          #
#	Stage the files listed in "files". The paths should have the 	#
#	'/pnfs/...' format. The pin lifetime is set with the value 	#
#	'srmv2_desiredpintime'. 						#
# ===================================================================== #
"""

from __future__ import print_function

import re
import sys
import gfal2 as gfal  # pylint: disable=import-error
from GRID_LRT.Staging import srmlist as srmlist_module


##>>> ctx = gfal2.creat_context()
#>>> (status, token) = ctx.bring_online(surl, 60, 60, True)


class LTA_Stager(object):
    """Stager that handles a file/srmlist and stages it. It can wait
    for the staging to complete"""

    def __init__(self,filename=None, srmlist=None):
        if srmlist and isinstance(srmlist,srmlist_module.srmlist):
            self.srmlist = srmlist
        else:
            self.srmlist = srmlist_module.srmlist()
        pass

    def load_from_filemane(self, filename):
        for i in open(filename):
            self.srmlist.append(i.strip('\r\n'))



def main(filename):
    """Given a filename, it stages the srms
    and prints 'staged' if completed
    """
    file_loc = location(filename)
    replace_string, match = replace(file_loc)
    srmfile = open(filename, 'r')
    urls = srmfile.readlines()
    srmfile.close()
    return process(urls, replace_string, match)


def state_dict(srm_dict):
    """Decides on the location of the data using
    a srm link and creates a dictionary
    """
    locs_options = ['s', 'j', 'p']

    line = srm_dict.itervalues().next()
    file_loc = [locs_options[i] for i in range(len(locs_options)) if [
        "sara" in line, "juelich" in line,
        "sara" not in line and "juelich" not in line][i] is True][0]
    replace_string, match = replace(file_loc)

    urls = []
    for _, value in srm_dict.iteritems():
        urls.append(value)
    return process(urls, replace_string, match)


def location(filename):
    """Gives the location of the entire filename
    """
    locs_options = ['s', 'j', 'p']
    with open(filename, 'r') as srmfile:
        line = srmfile.readline()

    file_loc = [locs_options[i] for i in range(len(locs_options)) if [
        "sara" in line, "juelich" in line, "sara" not in line and "juelich" not in line][i] is True]
    return file_loc[0]


def replace(file_loc):
    """Replaces the srmlink with the manager used to stage data"""
    if file_loc == 'p':
        match = re.compile('/lofar')
        repl_string = "srm://lta-head.lofar.psnc.pl:8443/srm/managerv2?SFN=/lofar"
        print("Staging in Poznan")
    else:
        match = re.compile('/pnfs')
        if file_loc == 'j':
            repl_string = "srm://lofar-srm.fz-juelich.de:8443/srm/managerv2?SFN=/pnfs/"
            print("Staging in Juleich")
        elif file_loc == 's':
            repl_string = "srm://srm.grid.sara.nl:8443/srm/managerv2?SFN=/pnfs"
            print("files are on SARA")
        else:
            sys.exit()
    return repl_string, match

def stage_srm(surl, pintime, timeout, asynch=True):
    context = gfal.creat_context()
    (errors, token) = context.bring_online(surl, pintime, timeout, asynch)
    return errors, token

def process(urls, repl_string, match):
    """Main function that invokes
    gfal on all the srms to stage them"""
    surls = []
    for url in urls:
        surls.append(url.strip())

    err, tok = stage_srm(surls, 5*24*3600, 3600)

    req = {}
    # Set the timeout to 24 hours
    # gfal_set_timeout_srm  Sets  the  SRM  timeout, used when doing an asyn-
    # chronous SRM request. The request will be aborted if it is still queued
    # after 24 hours.
    # Set the time that the file stays pinned on disk for a week (604800sec)

    print("staging request sent")
    return  err, tok 

if __name__ == '__main__':
    if len(sys.argv) == 2:
        sys.exit(main(sys.argv[1]))
    else:
        sys.exit(main('files'))



#!/usr/bin/env python
"""
Python module to check the state of files using gfal and return their locality
# ===================================================================== #
# author: Ron Trompert <ron.trompert@surfsara.nl>	--  SURFsara    #
# helpdesk: Grid Services <grid.support@surfsara.nl>    --  SURFsara	#
#                                                            	        #
# usage: python state.py						#
# description:                                                       	#
#	Display the status of each file listed in "files". The paths 	#
#	should have the '/pnfs/...' format. Script output:		#
#		ONLINE: means that the file is only on disk		#
#		NEARLINE: means that the file in only on tape		#
#		ONLINE_AND_NEARLINE: means that the file is on disk	#
#				     and tape				#
# ===================================================================== #
"""

from __future__ import print_function
import sys
from collections import Counter


try:
    import gfal2 as gfal  # pylint: disable=import-error
    gfal.set_verbose(gfal.verbose_level.warning)
except ImportError:
    print("GFAL CANNOT BE IMPORTED")
from GRID_LRT.Staging.srmlist import srmlist
from GRID_LRT.auth import grid_credentials



def main(filename, verbose=True):
    """Main function that takes in a file name and returns a list of tuples of
    filenames and staging statuses. The input file can be both srm:// and gsiftp:// links.

    Args:
        :param filename: The filename holding the links whose have to be checked
        :type filename: str
        :param verbose: A toggle to turn off printing out the status of each file.
        True by default will print everything out
        :type verbose: bool

    Returns:
        :ret results: A list of tuples containing the file_name and the State

    Usage:

    >>> from GRID_LRT.Staging import state_all
    >>> filename='/home/apmechev/GRIDTOOLS/GRID_LRT/GRID_LRT/tests/srm_50_sara.txt'
    >>> results=state_all.main(filename)
    >>> results=state_all.main(filename, verbose=False)
    >>> results[0]
    ('L229507_SB150_uv.dppp.MS_f6fc7fc5.tar', 'ONLINE_AND_NEARLINE')

    """
    grid_credentials.grid_credentials_enabled()  # Check if credenitals enabled
    s_list = load_file_into_srmlist(filename)
    print("files are at "+s_list.lta_location)
    results = []
    for i in s_list.gfal_links():
        results.append(check_status(i, verbose))
    return results


def load_file_into_srmlist(filename):
    """Helper function that loads a file into an srmlist object (will be
    added to the actual srmlist class later)

    """
    s_list = srmlist()
    for i in open(filename, 'r').read().split():
        s_list.append(i)
    return s_list


def check_status(surl_link, verbose=True):
    """ Obtain the status of a file from the given surl.

    Args:
        :param surl: the SURL pointing to the file.
        :type surl: str
        :parame verbose: print the status to the terminal.
        :type verbose: bool

    Returns:
        :(filename, status): a tuple containing the file  and status as
            stored in the 'user.status' attribute.
    """
    context = gfal.creat_context()
    status = context.getxattr(surl_link, 'user.status')
    filename = surl_link.split('/')[-1]
    if status == 'ONLINE_AND_NEARLINE' or status == 'ONLINE':
        color = "\033[32m"
    else:
        color = "\033[31m"
    if verbose:
        print('{:s} is {:s}{:s}\033[0m'.format(filename, color, status))
    return (filename, status.strip())


def percent_staged(results):
    """Takes list of tuples of (srm, status) and counts the percentage of files
    that are staged (0->1) and retunrs this percentage as float

    Usage:

    >>> from GRID_LRT.Staging import state_all
    >>> filename='/home/apmechev/GRIDTOOLS/GRID_LRT/GRID_LRT/tests/srm_50_sara.txt'
    >>> results=state_all.main(filename, verbose=False)
    >>> state_all.percent_staged(results)

    """
    total_files = len(results)
    counts = Counter(x[1] for x in results)
    staged = dict(counts).get('ONLINE_AND_NEARLINE', 0)+dict(counts).get('ONLINE', 0)
    unstaged = dict(counts).get('NEARLINE', 0)
#    assert staged + unstaged == total_files
    print(str(float(staged/(staged + unstaged))*100)+" percent of files staged")
    return float(staged/total_files)


def check_status_file(surl_list):  # TODO: implement
    """Unimplemented task"""
    print(surl_list)


if __name__ == '__main__':
    INPUT = sys.argv[1]
    if (INPUT.lower()).startswith('srm://'):
        # Single file.
        check_status(INPUT)
    else:
        # Assume a file list.
        check_status_file(INPUT)



#!/usr/bin/env python
"""Module to read and store PiCaS credentials

"""
from __future__ import print_function
from os.path import expanduser, isfile
from os import environ, chmod
from collections import namedtuple
import logging

__author__ = "Alexandar Mechev"
__credits__ = ["Alexandar Mechev", "Natalie Danezi", "Timothy Shimwell",
               "Raymond Oonk"]
__license__ = "GPL 3.0"
__version__ = "0.2.4"
__maintainer__ = "Alexandar Mechev"
__email__ = "LOFAR@apmechev.com"
__status__ = "Production"


def infolog(string):
    """infolog
    Logs to the info logger or else prints

    :param string: the string to log to the logger or to just print
    :type string: str
    """
    if not logging:
        print("INFO: "+string)
    else:
        logging.info(string)


def debuglog(string):
    """Logs to the debug logger or else prints"""
    if not logging:
        print("DEBUG: "+string)
    else:
        logging.debug(string)


def warnlog(string):
    """Logs to the warning logger or else prints"""
    if not logging:
        print("WARNING: "+string)
    else:
        logging.warning(string)

def get_picas_cred():
    Creds = namedtuple('Creds',['user','password','database'])
    pc = picas_cred()
    creds = Creds(pc.user, pc.password, pc.database)
    return  creds

#Will be renamed to PicasCred()
class picas_cred(object):
    """Getting PiCaS credentials from environment or .picasrc file

    This class stores the picas credential in internal variables:
        user: PiCaS username
        password: PiCaS password
        database: (default) PiCaS database

    The ~/.picasrc file is in the following format:

user=picas_username
password=picas_password
database=picas_database

    """

    def __init__(self, source_file=None, usr=None, pwd=None, dbn=None):
        """ picas_creds gets credentials either automatically or manually

        If custom variables file exists, credentials are taken from source_file
        default source file is ~/.picasrc, gets from there if file exists
        If no ~/.picasrc, the credentials are from environment
        Finally, credentials can be put in the constructor

        Args:
            source_file (str): the location of file storing PiCaS credentials
            usr (str): picas user with which to initialize class
            pwd (str): picas password for above user
            dbn (str): picas database
        """
        if source_file:
            self.get_picas_creds_from_file(pic_file=source_file)
        elif usr is None or pwd is None and dbn:
            if isfile(expanduser('~/.picasrc')):
                self.get_picas_creds_from_file('~/.picasrc')
            else:
                self.get_picas_creds_from_env()
        else:
            self.user = usr
            self.password = pwd
            self.database = dbn

    def get_picas_creds_from_file(self, pic_file='~/.picasrc'):
        """ loads picas credentials from a file (default: ~/.picasrc)

            Args:
                pic_file(str): file from which to load the credentials
                               (default: ~/.picasrc)
        """
        with open(expanduser(pic_file), 'r') as pc_file:
            infolog(str(
                "picas_credentials: Parsing user credentials from "+str(expanduser(pic_file))))
            for line in pc_file:
                if line.startswith("user"):
                    self.user = line.split('=')[1].strip()
                if line.startswith("password"):
                    self.password = line.split('=')[1].strip()
                if line.startswith("database"):
                    self.database = line.split('=')[1].strip()

    def get_picas_creds_from_env(self):
        """ Loads the PiCaS credentials from the environment

            The username, password and database must be stored
            in the variables $PICAS_USR, $PICAS_USR_PWD and $PICAS_DB

        """
        try:
            self.user = environ['PICAS_USR']
            self.password = environ['PICAS_USR_PWD']
            self.database = environ['PICAS_DB']
        except KeyError:
            warnlog("PICAS Variable not in ENV!")

    def get_picas_creds(self):
        """ Automatically gets the PiCaS credentials

            Note:
                This function preferably gets the variables from environment
        """
        if (not environ.get('PICAS_USR') and not environ.get('PICAS_USR_PWD') and
                not environ.get('PICAS_DB')):
            return self.get_picas_creds_from_env()
        return self.get_picas_creds_from_file()

    def put_picas_creds_in_env(self, picas_db=None):
        """ Inserts PiCaS credentials in environment

        This method is useful if scripts in this shell and subshells
        require the variables to be inside the environment.

        Args:
            picas_db (str): User can set a custom database, by default,
                            the one from the class is used
        """
        if picas_db:
            self.database = picas_db
            environ['PICAS_DB'] = picas_db
        environ['PICAS_USR'] = self.user
        environ['PICAS_USR_PWD'] = self.password
        if self.database:
            environ['PICAS_DB'] = self.database
        else:
            environ['PICAS_DB'] = self.database


    def put_creds_in_file(self, pic_file="~/.picasrc"):
        """put_creds_in_file
            Exports PiCaS variables into a file

        :param pic_file: Name of the file containing the credentials (optional)
        :type pic_file: str

        """
        with open(expanduser(pic_file), 'w') as pc_file:
            pc_file.write("user="+str(self.user)+"\n")
            pc_file.write("password="+str(self.password)+"\n")
            pc_file.write("database="+str(self.database)+"\n")
        chmod(expanduser(pic_file), 384)

    def return_credentials(self):
        """return_credentials
           Returns a dictionary of PiCaS credentials.

            d=pcred.return_credentials
            d.keys()=['user','password','database']
        """
        return {'user': self.user, 'password': self.password, 'database': self.database}


if __name__ == '__main__':  # code to execute if called from command-line
    infolog("PiCaS_creds from main()")
    if isfile(expanduser('~/.picasrc')):
        PCREDS = picas_cred(source_file='~/.picasrc')
        PCREDS.put_picas_creds_in_env()
    else:
        PCREDS = picas_cred()
        PCREDS.put_creds_in_file()



import datetime
from GRID_LRT.storage.gsifile  import GSIFile



to_delete = []

def get_files_in_dir(base_dir='gsiftp://gridftp.grid.sara.nl:2811/pnfs/grid.sara.nl/data/lofar/user/sksp/diskonly/pipelines/SKSP/prefactor_v1.0/pref_targ1/'):
    b_dir = GSIFile(base_dir)
    files = b_dir.list_dir()
    return files

def list_files_older_than(files, num_days=31):
    to_delete = []
    for f in files:
        if f.datetime < datetime.datetime.now() - datetime.timedelta(days=num_days):
            to_delete.append(f)
    return to_delete

def text_format(fileobj):
    if fileobj.is_file:
        preamble = "File {}".format(fileobj.filename)
    else:
        preamble = "Folder {} with {} Files".format(fileobj.filename, fileobj._get_num_files())
    text = "{} modified on {} will be deleted. ".format(preamble, fileobj.datetime.strftime("%Y-%m-%d"))
    return text


def print_message_for_folder(base_dir, numdays, message=None, **kwargs):
    files = get_files_in_dir(base_dir)
    todeletefiles = list_files_older_than(files, num_days=numdays)
    if not message:
        print("Here are all the files I'll delete on Monday")
    else:
        print(message)
    print("for folder {}".format(base_dir))
    for f in todeletefiles:
        print(text_format(f))

def cleanup_folder(base_dir, numdays):
    print_message_for_folder(base_dir, numdays, message="Deleting these files!")
    files = get_files_in_dir(base_dir)
    todeletefiles = list_files_older_than(files, num_days=numdays)
    for f in todeletefiles:
        if f.is_dir:
            for subdir in f.list_dir():
                subdir.delete()
            f.delete()
        else:
            f.remove()



# ===================================================================== #
# author: Ron Trompert <ron.trompert@surfsara.nl>	--  SURFsara    #
# helpdesk: Grid Services <grid.support@surfsara.nl>    --  SURFsara    #
#                                                                       #
# usage: python stage.py                                                #
# description:                                                          #
#	Stage the files listed in "files". The paths should have the 	#
#	'/pnfs/...' format. The pin lifetime is set with the value 	#
#	'srmv2_desiredpintime'. 					#
# ===================================================================== #

#!/usr/bin/env python

#import gfal2 as gfal
import re
import sys
from GRID_LRT.Staging import stager_access 


def strip(item):
    return item.strip()


def process_surl_line(line):
    """ Used to drop empty lines and to
        take the first argument of the srmfile (the srm:// link)
    """

    if " " in line:
        line = line.split(" ")[0]
    if line == "/n":
        return None 
    return line


def main(filename, test=False):
    file_loc = location(filename)
    rs, m = replace(file_loc)
    with open(filename, 'r') as f:
        urls = f.readlines()
    return process(urls, rs, m, test)


def return_srmlist(filename):
    file_loc = location(filename)
    regex, match = replace(file_loc)
    _file = open(filename, 'r')
    urls = _file.readlines()
    _file.close()
    surls = []
    for url in urls:
        url = process_surl_line(url)
        if "managerv2?SFN" in url:
            surls.append(match.sub(regex, strip(u)))
        elif url:
            surls.append(url)
    return surls


def state_dict(srm_dict):
    locs_options = ['s', 'j', 'p']

    line = srm_dict.itervalues().next()
    file_loc = [locs_options[i] for i in range(len(locs_options)) if [
        "sara" in line, "juelich" in line, not "sara" in line and not "juelich" in line][i] == True][0]
    regex, match = replace(file_loc)

    urls = []
    for key, value in srm_dict.iteritems():
        urls.append(value)
    return process(urls, regex, match)


def location(filename):
    locs_options = ['s', 'j', 'p']
    with open(filename, 'r') as _file:
        line = _file.readline()

        file_loc = [locs_options[i] for i in range(len(locs_options)) if [
            "sara" in line, "juelich" in line, not "sara" in line and not "juelich" in line][i] == True]
    return file_loc[0]


def replace(file_loc):
    if file_loc == 'p':
        match = re.compile('/lofar')
        repl_string = "srm://lta-head.lofar.psnc.pl:8443/srm/managerv2?SFN=/lofar"
        print("Staging in Poznan")
    else:
        match = re.compile('/pnfs')
        if file_loc == 'j':
            repl_string = "srm://lofar-srm.fz-juelich.de:8443/srm/managerv2?SFN=/pnfs/"
            print("Staging in Juleich")
        elif file_loc == 's':
            repl_string = "srm://srm.grid.sara.nl:8443/srm/managerv2?SFN=/pnfs"
            print("files are on SARA")
        else:
            sys.exit()
    return repl_string, match


def process(urls, repl_string, match, test=False):
    surls = []
    for url in urls:
        if not 'srm' in url:
            surls.append(match.sub(repl_string, strip(url)))
        else:
            surls.append(strip(url))
    req = {}
    print("Setting up "+str(len(surls))+" srms to stage")
    if test:
        return
    stageid = stager_access.stage(surls)

    print("staged with stageID ", stageid)
    return stageid


def get_stage_status(stageid):
    return stager_access.get_status(int(stageid))


if __name__ == '__main__':
    if len(sys.argv) == 2:
        sys.exit(main(sys.argv[1]))
    else:
        sys.exit(main('files'))



from datetime import datetime
import warnings
from subprocess import Popen, PIPE
import re
import sys
import GRID_LRT.auth.grid_credentials as grid_creds
from GRID_LRT.auth.get_picas_credentials import get_picas_cred
from GRID_LRT import token
from GRID_LRT.Staging.srmlist import srmlist
from GRID_LRT.storage.gsifile import GSIFile


def get_srmdir_from_token_task(token_type, view, key = 'RESULTS_DIR'):
    """Creates a list of files from a set of tokens. Returns a GSIFile object"""
    pc=get_picas_cred()
    th=Token.Token_Handler(t_type=token_type, uname=pc.user, pwd=pc.password, dbn=pc.database)
    tokens=th.list_tokens_from_view(view)
    srmdir, OBSID, pipeline_step = None, None, None
    for t in tokens: #TODO: Do this with a proper view
        if not pipeline_step:
            pipeline_step = th.database[t['id']]['pipeline_step']
            OBSID = th.database[t['id']]['OBSID']
        if srmdir is not None: break
        if OBSID:
            srmdir_location = str(th.database[t['id']][key])+"/"+pipeline_step+"/"+str(OBSID)
            srmdir = GSIFile(srmdir_location)
    return srmdir

def make_srmlist_from_srmdir(srmdir):
    slist = srmlist()
    for i in srmdir.list_dir():
        slist.append(i.location)
    return slist




#!/usr/bin/env python
"""Setup script for the Grid LOFAR Tools (GRID_LRT) python module
"""
#from distutils.core import setup
from setuptools import setup
import GRID_LRT
 
setup(name='GRID_LRT',
      packages=['GRID_LRT', 'GRID_LRT/Staging', 'GRID_LRT/application','GRID_LRT/storage', 'GRID_LRT/auth'],
      version=GRID_LRT.__version__,
      setup_requires=[],
      tests_require=[
          'pytest'
          ],
      install_requires=[
          'retrying',
          'cloudant',
          'humanfriendly' 
          ],
      include_package_data=True,
      data_files=[("GRID_LRT/data/config",
                   ["GRID_LRT/data/config/steps/cal_pref3.json",
                    "GRID_LRT/data/config/steps/targ1_pref3.json",
                    "GRID_LRT/data/config/steps/targ2_pref3.json",
                    "GRID_LRT/data/config/steps/pref_targ2.cfg"]),
                  ("GRID_LRT/data/launchers/",
                   ["GRID_LRT/data/launchers/run_remote_sandbox.sh"])
                 ],
      description='GRID LOFAR Reduction Tools',
      long_description="Software that encapsulates LOFAR processing"
      "and allows the use of High Throughput Distributed processing available"
      "at SURFsara and other European Grid Initiative locations.",
      author='Alexandar Mechev',
      author_email='apmechev+LOFAR@gmail.com',
      url='https://www.github.com/apmechev/GRID_LRT/',
      download_url='https://github.com/apmechev/GRID_LRT/archive/v{}.tar.gz'.format(GRID_LRT.__version__),
      keywords=['surfsara', 'distributed-computing', 'LOFAR'],
      classifiers=['Development Status :: 5 - Production/Stable',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Science/Research',
                   "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
                   "Natural Language :: English",
                   "Topic :: Scientific/Engineering :: Astronomy",
                   "Topic :: System :: Distributed Computing", 
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3.3',
                   'Programming Language :: Python :: 3.4', 
                   'Programming Language :: Python :: 3.5',
                   'Programming Language :: Python :: 3.6',
                   'Programming Language :: Python :: 3.7']
     )




#!/usr/bin/env python

"""This module creates the launch scripts for the WMS on gina
Options include the number of cores and the queue type."""

import os
import signal
import subprocess
import sys
from subprocess import Popen
import logging
import warnings
import random, string
from shutil import copyfile, rmtree


import tempfile
from GRID_LRT.auth.get_picas_credentials import picas_cred as pc
import GRID_LRT
from GRID_LRT.auth import grid_credentials

class SafePopen(Popen):
    def __init__(self, *args, **kwargs):
        if sys.version_info.major == 3 :
            kwargs['encoding'] = 'utf8'
        return super(SafePopen, self).__init__(*args, **kwargs)

 

#class job_launcher(object):
#    """Generic Job Launcher
#    """
#    def __init__(self):
#        pass

class gridjob(object):
    """A class containing all required descriptions for a SINGLE 
    Grid job, as well as its parameters"""
    def __init__(self, wholenodes=False, NCPU=1, token_type=None):
        self.token_type = token_type
        if wholenodes:
            self.wholenodes = 'true'
        else:
            self.wholenodes = 'false'
        self.ncpu = NCPU       



class RunningJob(object):
    def __init__(self, glite_url=''):
        self.job_status='Unknown'
        self.glite_status='Unknown'
        if glite_url:
            self.glite_url = glite_url
   
    @property
    def status(self):
        self.__check_status()
        return self.job_status

    def __check_status(self):
        glite_process = SafePopen(['glite-wms-job-status', self.glite_url],stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        result, err = glite_process.communicate()
        try:
            self.job_status=result.split('Current Status:')[1].split()[0]
        except:
            print(err)

        if self.glite_status== 'Running':
            self.count_successes(result)
        if self.glite_status=='Waiting':
            self.count_successes(result)
        if self.glite_status == 'Aborted':
            self.count_successes(result)
        if self.glite_status=='Running' and self.job_status=='Waiting':
            self.glite_status='Completed'
            
    def count_successes(self,jobs):
        """Counts the number of Completed jobs in the results of the glite-wms-job-status
        output.  """
        exit_codes=[]
        jobs_list=[]
        for j in jobs.split('=========================================================================='):
            jobs_list.append(j)
        statuses=[]
        for j in jobs_list:
            if "Current Status:" in j:
                statuses.append(j.split("Current Status:")[1].split('\n')[0])
        numdone=0
        for i in statuses:
            if 'Done' in i or 'Cancelled' in i or 'Aborted' in i  :
                numdone+=1
        if 'Done' in statuses[0] or 'Aborted' in statuses[0]:
            self.job_status = 'Done'
        if numdone == len(jobs_list):
            self.job_status='Done'
        if self.job_status == 'Waiting':
            for i in statuses:
                if 'Scheduled' in i:
                    self.job_status = 'Scheduled'
                elif 'Running' in i: 
                    self.job_status = 'Running'
                elif 'Submitted' in i:
                    self.job_status = 'Submitted'
                else:
                    self.job_status = "Done"
        logging.info("Num_jobs_done "+str(numdone)+" snd status is "+self.job_status)
        self.numdone = numdone
        return statuses[1:]

    def __str__(self):
        return "<Grid job '{}' with status '{}'>".format(self.glite_url, self.status)

    def __repr__(self):
        return self.__str__()


class JdlLauncher(object):
    """jdl_launcher creates a jdl launch file with the
    appropriate queue, CPUs, cores and Memory per node

    The jdl file is stored in a temporary location and can be
    automatically removed using a context manager such as:
    >>> with Jdl_launcher_object as j:
        launch_ID=j.launch()
    This will launch the jdl, remove the temp file and reutrn the
    Job_ID for the glite job.
    """

    def __init__(self, numjobs=1, token_type='t_test',
                 parameter_step=4, **kwargs):
        """The jdl_launcher class is initialized with the number of jobs,
        the name of the PiCaS token type to run, and a flag to use the whole node.


         Args:
            numjobs (int): The number of jobs to launch on the cluster
            token_type (str): The name of the token_type to launch from the
                PiCaS database. this uses the get_picas_credentials module to
                get the PiCaS database name, uname, passw
            wholenodes(Boolean): Whether to reserve the entire node. Default is F
            NCPU (int, optional): Number of CPUs to use for each job. Default is 1
        """

        self.authorized = False
        if 'authorize' in kwargs.keys() and kwargs['authorize'] == False:
                warnings.warn("Skipping Grid Autorization")
        else: 
	        self.__check_authorized()
        if numjobs < 1:
            logging.warn("jdl_file with zero jobs!")
            numjobs = 1
        self.numjobs = numjobs
        self.parameter_step = parameter_step
        self.token_type = token_type
        self.wholenodes = 'false'

        if 'wholenode' in kwargs:
            self.wholenodes = kwargs['wholenode']
        if "NCPU" in kwargs:
            self.ncpu = kwargs["NCPU"]
        else:
            self.ncpu = 1
        if self.ncpu == 0:
            self.wholenodes = 'true'
        if "queue" in kwargs:
            self.queue = kwargs['queue']
        else:
            self.queue = "medium"
        self.temp_file = None
        self.launch_file = str("/".join((GRID_LRT.__file__.split("/")[:-1])) +
                               "/data/launchers/run_remote_sandbox.sh")

    def __check_authorized(self):
        grid_credentials.grid_credentials_enabled()
        self.authorized = True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        os.remove(self.temp_file.name)

    def build_jdl_file(self, database=None):
        """Uses a template to build the jdl file and place it in a
        temporary file object stored internally.
            """
        creds = pc()  # Get credentials, get launch file to send to workers
        if not database:
            database = creds.database
        if not os.path.exists(self.launch_file):
            raise IOError("Launch file doesn't exist! "+self.launch_file)
        jdlfile = """[
  JobType="Parametric";
  ParameterStart=0;
  ParameterStep=%d;
  Parameters= %d ;
  Executable = "/bin/sh";

  Arguments = "run_remote_sandbox.sh %s %s %s %s ";
  Stdoutput = "parametricjob.out";
  StdError = "parametricjob.err";
  InputSandbox = {"%s"};
  OutputSandbox = {"parametricjob.out", "parametricjob.err"};
  DataAccessProtocol = {"gsiftp"};
  ShallowRetryCount = 0;

  Requirements=(RegExp("gina.sara.nl:8443/cream-pbs-%s",other.GlueCEUniqueID));
  WholeNodes = %s ;
  SmpGranularity = %d;
  CPUNumber = %d;
]""" % (int(self.parameter_step),
        int(self.numjobs),
        str(database),
        str(creds.user),
        str(creds.password),
        str(self.token_type),
        str(self.launch_file),
        str(self.queue),
        str(self.wholenodes),
        int(self.ncpu),
        int(self.ncpu))
        return jdlfile

    def make_temp_jdlfile(self, database=None):
        """ Makes a temporary file to store the JDL
        document that is only visible to the user"""
        self.temp_file = tempfile.NamedTemporaryFile(delete=False)
        print("making temp file at "+self.temp_file.name)
        with open(self.temp_file.name, 'w') as t_file_obj:
            for i in self.build_jdl_file(database):
                t_file_obj.write(i)
        return self.temp_file

    def launch(self, database=None):
        """Launch the glite-job and return the job identification"""
        if not self.authorized:
	        self._check_authorized()
        if not self.temp_file:
            self.temp_file = self.make_temp_jdlfile(database = database)
        sub = safePopen(['glite-wms-job-submit', '-d', os.environ["USER"],
                                self.temp_file.name], stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
        out = sub.communicate()
        if out[1] == "":
            return out[0].split('Your job identifier is:')[1].split()[0]
        raise RuntimeError("Launching of JDL failed because: "+out[1])



class UnauthorizedJdlLauncher(JdlLauncher):
    def __init__(self, *args, **kw):
        super(UnauthorizedJdlLauncher, self).__init__(*args, authorize=False, **kw)

    def launch(self):
        if not self.temp_file:
            self.temp_file = self.make_temp_jdlfile()
        fake_link = 'https://wms2.grid.sara.nl:9000/'+''.join(random.choice(string.ascii_letters + string.digits+"-") for _ in range(22))
        warnings.warn("If you were authorized, we would be launching the JDL here. You'd get this in return: {}".format(fake_link))
        return fake_link


class LouiLauncher(JdlLauncher):
    """
        To make integration tests of an AGLOW step, this job launches it on loui.
    """

    def __init__(self, *args, **kwargs):
        super(LouiLauncher,self).__init__(*args, **kwargs)
        self.pid=None
        self.return_directory = os.getcwd()
        self.run_directory = tempfile.mkdtemp(prefix='/scratch/') 
        
    def launch(self, database=None):
        copyfile(self.launch_file, self.run_directory+"/run_remote_sandbox.sh")
        os.chdir(self.run_directory)
        creds = pc()
        if not database:
            database = creds.database
        command = "./run_remote_sandbox.sh {} {} {} {}".format(database,
                            creds.user, creds.password, self.token_type)
        os.chmod('run_remote_sandbox.sh',0o744)
        print("Running in folder: ")
        print("")
        print("Don't forget to run LouiLauncher.cleanup() in Pythonwhen you're done!")
        print(self.run_directory)
        with open(self.run_directory+"/stdout.txt","wb") as out:
            with open(self.run_directory+"/stderr.txt","wb") as err:
                launcher = SafePopen(command.split(), stdout=out, stderr=err)
                self.pid = launcher.pid
                launcher.wait()
        return {'output':self.run_directory+"/stdout.txt",
                'error':self.run_directory+"/stderr.txt"}

    def __check_authorised(self):
        self.authorized = True
    
    def cleanup(self):
        print("removing directory " + self.run_directory)
        rmtree(self.run_directory)
        os.chdir(self.return_directory)
        if self.pid:
            os.kill(self.pid, signal.SIGKILL)

    def __del__(self):
        if os.path.exists(self.run_directory):
            self.cleanup()

    def __exit__(self, exc_type, exc_value, traceback):
        return None



# -*- coding: utf-8 -*-
#
# Copyright (C) 2016 Alexandar P. Mechev
# All rights reserved.
#
# This software is licensed as described in the file LICENSE.md, which
# you should have received as part of this distribution.

"""
.. module:: GRID_LRT.Token
   :platform: Unix
   :synopsis: Set of tools for manually and automatically creating tokens

.. moduleauthor:: Alexandar Mechev <LOFAR@apmechev.com>
.. note:: Will be renamed GRID_LRT.token to conform to python standards

>>> #Example creation of a token of token_type 'test'
>>> from GRID_LRT.auth.get_picas_credentials import picas_cred
>>> pc=picas_cred() #Gets picas_credentials
>>>
>>> from cloudant.client import CouchDB         
>>> from GRID_LRT.token import caToken #Token that implements the cloudant interface
>>> from GRID_LRT.token import TokenList
>>> client = CouchDB(pc.user,pc.password, url='https://picas-lofar.grid.surfsara.nl:6984',connect=True)
>>> db = client[pc.database]
>>> 'token_id' in db # Checks if database includes the token
>>> db['token_id'] #Pulls the token 
>>> tl = TokenList(database=db, token_type='token_type') #makes an empty list
>>> tl.add_view(TokenView('temp',"doc.type == \"{}\" ".format(tl.token_type))) # adds a view to the token
>>> for token in tl.list_view_tokens('temp'):
        tl.append(caToken(token_type=tl.token_type, token_id= token['_id'], database=db))
        # adds every token in view to the local list (only their ids)
>>> tl.fetch() #Fetch actual data for token in list
>>> t1 = caToken(database=db, token_type='restructure_test', token_id=token_id) #make a token (locally)
>>> t1.build(TokenJsonBuilder('/path/to/token/data.json'))
>>> t1.save() #upload to the database
>>> t1.add_attachment(attachment_name='attachment_name_in_db.txt',filename='/path/to/attachment/file') #Adds attachment to token
"""

from __future__ import print_function
import sys
import os
import shutil
import itertools
import time
import tarfile
import json

# from retrying import retry
from abc import ABCMeta, abstractmethod
import pdb

import mimetypes
import GRID_LRT
from cloudant import couchdb as cloudant_couchdb
from cloudant.design_document import DesignDocument
from cloudant import couchdb_admin_party

# from cloudant.client import CouchDB
from cloudant import couchdb as CaCouchClient
from cloudant.document import Document
from cloudant.error import CloudantArgumentError
from requests.exceptions import HTTPError

__version__ = GRID_LRT.__version__
__author__ = GRID_LRT.__author__
__license__ = GRID_LRT.__license__
__email__ = GRID_LRT.__email__
__copyright__ = GRID_LRT.__copyright__
__credits__ = GRID_LRT.__credits__
__maintainer__ = GRID_LRT.__maintainer__
__status__ = GRID_LRT.__status__


def get_all_design_docs(pcreds=None, srv="https://picas-lofar.grid.surfsara.nl:6984"):
    """Returns a list of design documents for the pcreds.databse on server=srv
    If pcreds are none, then we're adminparty and db is test_db"""
    kwargs = {"connect": True, "url": srv}
    if pcreds:
        user, passwd, dbn = pcreds.user, pcreds.password, pcreds.database
        connect_client = CaCouchClient
        kwargs["user"] = user
        kwargs["passwd"] = passwd
    else:
        user, passwd, dbn = None, None, "test_db"
        connect_client = couchdb_admin_party
    with connect_client(**kwargs) as client:
        database = client[dbn]
        ad = [doc for doc in database["_all_docs"]["rows"] if "_design" in doc["id"]]
    return [i["id"] for i in ad]


def reset_all_tokens(
    token_type, picas_creds, server="https://picas-lofar.grid.surfsara.nl:6984"
):
    """ Resets all Tokens with the pc authorization
    """
    thandler = TokenHandler(
        t_type=token_type,
        srv=server,
        uname=picas_creds.user,
        pwd=picas_creds.password,
        dbn=picas_creds.database,
    )
    thandler.load_views()
    for view in list(thandler.views):
        if view != "overview_total":
            thandler.reset_tokens(view)


def purge_tokens(
    token_type, picas_creds, server="https://picas-lofar.grid.surfsara.nl:6984"
):
    """Automated function to purge tokens authorizing with Picas_creds"""
    thandler = TokenHandler(
        t_type=token_type,
        srv=server,
        uname=picas_creds.user,
        pwd=picas_creds.password,
        dbn=picas_creds.database,
    )
    thandler.load_views()
    thandler.purge_tokens()


class Token(dict):
    def __init__(self, token_type, token_id=None, **kwargs):
        """Base token class, that inherits from a python dict. You can access token fields
        using the python __get__ method, i.e. token['field']. Each token has a token_id that is unique
        in the database, which it is part of, however the database interface is only defined in the 
        child classes of Token"""
        self.__setitem__("type", token_type)
        if not token_id:
            self.__setitem__("_id", token_type)
        else:
            self.__setitem__("_id", token_id)
        self.__setitem__("PICAS_API_VERSION", GRID_LRT.__version__)
        self.__setitem__("lock", 0)
        self.__setitem__("done", 0)

    @property
    def database(self):
        return self._database

    @database.setter
    def database(self, db):
        self._database = db
    
    def synchronize(self, db, prefer_local=False, upload=False):
        """Synchronizes the token with the database. This method requires
        the child class of Token to have a .upload(database) interface available
        """
        remote_token = db[self["_id"]]
        for k in set(list(remote_token.keys()) + list(self.keys())):
            if prefer_local:
                remote_token[k] = self.get(k)
            else:
                self[k] = remote_token.get(k)
        if upload:
            self.upload(db)

    def build(self, token_builder):
        """Uses a TokenBuilder to load data into the token. TokenBuilders can
        interface Tokens with different ways to build them (i.e. from a dict in pyton,
        from json files, from streams, etc)"""
        data = token_builder.data
        self.update(data)

    def add_attachment(self):
        raise NotImplementedError

    def reset(self):
        """General interface to 'reset' a Token as defined in PiCaS Standards. 
        The lock and done fields are timestamps and if are set to 0, the token will be
        considered in the 'todo' state"""
        self.__setitem__("lock", 0)
        self.__setitem__("done", 0)
        scrub_count = self.get(
            "scrub_count", 0
        )  ##TODO: Get function override to get from db
        self.__setitem__("scrub_count", scrub_count + 1)
        self.__setitem__("hostname", "")
        self.__setitem__("output", "")
        self.__setitem__("status", "reset")

    def archive(self, delete=False):
        """Archives all tokens, given a database. Deletes them if the delete flag is set to True"""
        self.get_all_attachments()
        json.dump(self, open(self._filename() + ".json", "wb"))
        if delete:
            db = self.database
            del db[self["_id"]]

    def _filename(self):
        filename = self["_id"]
        filename = filename.replace("/", "_")
        return filename


class caToken(Token, Document):
    def __init__(self, database, token_type, **kwargs):
        """The caToken is a Token implementation using the python cloudant client
        to store tokens in a remote couchDB database. """
        Token.__init__(self, token_type=token_type, **kwargs)
        Document.__init__(self, database=database, **kwargs)
        self._document_id = self["_id"]

    def add_attachment(self, filename, attachment_name):
        file_type = mimetypes.guess_type(filename)[0]
        if "." in filename and filename.split(".")[-1] == "parset":
            file_type = "text/plain"
        self.put_attachment(attachment_name, file_type, open(filename, "r"))

    def get_all_attachments(self):
        """Gets all attachments from the remote token and saves them
        to a file with name ID-Attachment"""
        self.fetch()
        for attachment in self.get("_attachments"):
            if not self["_attachments"][attachment].get("content-type"):
                data = self.get_attachment(attachment, attachment_type="text/plain")
            else:
                data = self.get_attachment(attachment)
            attachment = attachment.replace("/", "_")
            with open(self._filename() + "-" + attachment, "wb") as att_f:
                att_f.write(data)


class TokenBuilder:
    __metaclass__ = ABCMeta
    """Creates a token"""

    @abstractmethod
    def _build(self):
        pass

    @property
    def data(self):
        return self._data


class TokenDictBuilder(TokenBuilder):
    def __init__(self, config_dict):
        self._build(config_dict)

    def _build(self, config_dict):
        self._build_from_dict(config_dict)

    def _build_from_dict(self, config_dict):
        self._data = {}
        if "PicasApiVersion" in config_dict and config_dict["PicasApiVersion"] < 0.5:
            raise RuntimeError(
                "Unsupported PiCaS API version {0}", config_dict["PicasApiVersion"]
            )
        self._data["config.json"] = {}
        _config = config_dict
        _variables = {}

        if "Token" in _config:
            if "variables" in _config["Token"]:
                _variables.update(_config["Token"]["variables"])
                _config["variables"] = _config["Token"]["variables"]
                del _config["Token"]["variables"]
            self._data.update(_config["Token"])

        if "Job" in _config and "variables" in _config["Job"]:
            self._data.update(_config["Job"]["variables"])

        if "variables" in _config:
            self._data["config.json"]["variables"] = _config["variables"]
        if "container" in _config:
            self._data["config.json"]["container"] = _config["container"]
        if "sandbox" in _config:
            self._data["config.json"]["sandbox"] = _config["sandbox"]


class TokenJsonBuilder(TokenDictBuilder):
    """Reads a json config file and builds the 
    Token fields using the data in this file."""

    def __init__(self, config_file):
        self._build(config_file)

    def _build(self, config_file):
        _config = json.load(open(config_file))
        self._data = {}
        self._build_from_dict(_config)


class TokenList(list):
    """A list of token that a Token can be appended to
    Includes upload and download functions to upload all the 
    local documents and to download the remote ones
    
    NOTE: Implement this in a composite pattern, to allow sublists
    where each sublist is the result of a view"""

    def __init__(self, token_type=None, database=None):
        self._token_ids = []
        self._design_doc = None
        self.database = database
        self.__set_token_type(token_type)

    @property
    def database(self):
        return self._database

    def __setitem__(self,key, val):
        for item in self:
            item.__setitem__(key,val)
            item.save()

    def __set_token_type(self, token_type):
        self.token_type = token_type
        if token_type and self._database != None:
            self.__set_design_doc()

    @database.setter
    def database(self, database):
        if not hasattr(self, 'database'):
            self._database = database


    def __set_design_doc(self):
        """This sets the internal variable _design_doc, so that we can reference
        the tokenList's design document for updates such as views. Also, we
        include the GRID_LRT version inside the ddoc, since we'll use the same version
        of the PiCaS_Launcher to process our jobs"""
        design_doc_name = "_design/{0}".format(self.token_type)
        self._design_doc = DesignDocument(self.database, design_doc_name)
        if design_doc_name not in self.database:
            self._design_doc.save()
        self._design_doc.fetch()
        self._design_doc.__setitem__("PICAS_API_VERSION", GRID_LRT.__version__)
        self._design_doc.save()

    def add_attachment(self, filename, attachment_name):
        for token in self:
            try:
                token.add_attachment(filename, attachment_name)
            except HTTPError as e:
                print(e)
                if "404" in str(e):
                    token.save()
                    token.add_attachment(filename, attachment_name)

    def append(self, item):
        if isinstance(item, Token):
            if not hasattr(self, 'database'):
                self.database=item.database  
            if not hasattr(self, 'token_type'):
                self.__set_token_type(item["type"])
            elif item["type"] != self.token_type:
                raise TypeError(
                    "Appending token of wrong token type, {0} != {1}".format(
                        item["type"], self.token_type
                    )
                )
            if item["_id"] not in self._token_ids:
                self._token_ids.append(item["_id"])
            else:
                raise RuntimeError(
                    "token with id {0} already exists! You need unique '_id' fields ".format(
                        item["_id"]
                    )
                )
            super(TokenList, self).append(item)
        elif isinstance(item, TokenList):
            for _id in item._token_ids:
                self._token_ids.append(_id)
            super(TokenList, self).append(item)
        else:
            raise TypeError("Cannot append item {0} as it's not a Token".format(item))

    def upload_all(self):
        """Uploads all tokens in the list. It will crash if the token already exists
        This also cannot be called if the tokens were created inside a context manager
        that has been exited. Best option is to create a persistent 'cloudant.client.CouchDB'
        object. """
        for token in self:
            token.save()

    def get_all_remote_tokens(self):
        for view in self.get_views():
            if view != "overview_total":
                for token in self.list_view_tokens(view):
                    if token not in self:
                        self.append(token)

    def save(self):
        self.upload_all()

    def delete_all(self):
        "Deletes all tokens inside the token list, including recursive lists"
        for token in self:
            token.delete()

    def reset(self):
        """Calls the reset method of all objects in the TokenList (either Tokens
        or sub-lists)"""
        for token in self:
            token.reset()

    def delete(self):
        """Just a wrapper around delete_all to standardize the token.delete() interface"""
        self.delete_all(self)

    def add_view(self, view):
        """Adds a view to the tokenList, and its respective design document. 
        the View object needs to return a map code and reduce code"""
        map_code = view.get_codes(self.token_type)[0]
        reduce_code = view.get_codes(self.token_type)[1]
        self._design_doc.fetch()
        if self._design_doc:
            try:
                self._design_doc.add_view(view.name, map_code, reduce_code)
            except CloudantArgumentError:
                self._design_doc.delete_view(view.name)
                self._design_doc.add_view(view.name, map_code, reduce_code)
        self._design_doc.save()

    def fetch(self):
        """For each item in the TokeList, we call 'fetch' """
        for token in self:
            token.fetch()
        self._design_doc.fetch()

    def list_view_tokens(self, view_name):
        view = self._design_doc.get_view(view_name)
        if not view or "reduce" in view:
            return self
        view_list = TokenList(token_type=self.token_type, database=self.database)
        for i in view.result:
            tok = caToken(
                database=self.database, token_type=self.token_type, token_id=i["id"]
            )
            view_list.append(tok)
        view_list.fetch()
        return view_list

    def get_views(self):
        """Returns all the views in the token's Design_doc"""
        if self._design_doc:
            self._design_doc.fetch()
            return self._design_doc.list_views()

    def delete_views(self):
        """Deletes all views listed in the token's DesignDocument"""
        for view in self.get_views():
            self._design_doc.delete_view(view)
        self._design_doc.save()

    def delete_ddoc(self):
        """Deletes the tokenList's designdocument from the database"""
        ddoc = self.database.get_design_document(self.token_type)
        ddoc.delete()

    def archive(self, compress=False, delete=False):
        """Archives all tokens in the tokenlist"""
        curr_dir = os.getcwd()
        save_dir = self._design_doc["_id"].split("/")[1]
        if not os.path.exists(save_dir):
            os.mkdir(save_dir)
        os.chdir(save_dir)
        for token in self:
            token.archive(delete=delete)
        if compress:
            with tarfile.open("{0}.tar.gz".format(save_dir), "w:gz") as tf:
                for savefile in os.listdir(os.getcwd()):
                    tf.add(savefile)
        self.delete_ddoc()
        os.chdir(curr_dir)

    def add_token_views(self):
        """Adds the todo, locked, error, done and overview_view 
        views that are standard for a PiCaS Token."""
        self.add_view(TokenView("todo", "doc.lock ==  0 && doc.done == 0 "))
        self.add_view(
            TokenView(
                "locked", "doc.lock > 0 && doc.done == 0 ", ("doc._id", "doc.status")
            )
        )
        self.add_view(TokenView("done", 'doc.status == "done"'))
        self.add_view(
            TokenView("error", 'doc.status == "error" ', ("doc._id", "doc.status"))
        )
        self.add_view(TokenReduceView("overview_view"))


class TokenView(object):
    def __init__(self, name, condition, emit_values=("doc._id", "doc._id")):
        self.name = name
        self.condition = condition
        self.emit_values = emit_values

    def _get_map_code(self, token_type):
        general_view_code = """function(doc) {{
   if(doc.type == "{0}") {{
      if({1}) {{
         emit({2}, {3});
      }} 
   }} 
}}
""".format(
            token_type, self.condition, self.emit_values[0], self.emit_values[1]
        )
        return general_view_code

    def _get_reduce_code(self):
        return None

    def get_codes(self, token_type):
        return self._get_map_code(token_type), self._get_reduce_code()


class TokenReduceView(TokenView):
    def __init__(self, name, **kwargs):
        super(TokenReduceView, self).__init__(name, kwargs)

    def _get_map_code(self, token_type):
        overview_map_code = """function(doc) {{
   if(doc.type == "{0}" )
     {{
       if (doc.lock == 0 && doc.done == 0){{
          emit('todo', 1);
       }}
       if(doc.lock > 0 && doc.status == 'downloading' ) {{
          emit('downloading', 1);
       }}
       if(doc.lock > 0 && doc.done > 0 && doc.output == 0 ) {{
          emit('done', 1);
       }}
       if(doc.lock > 0 && doc.output != 0 && doc.output != "" ) {{
          emit('error', 1);
       }}
       if(doc.lock > 0 && doc.status == 'launched' ) {{
          emit('waiting', 1);
       }}
       if(doc.lock > 0  && doc.done==0 && doc.status!='downloading' ) {{
          emit('running', 1);
      }}
     }}  
}}
""".format(
            token_type
        )
        return overview_map_code

    def _get_reduce_code(self):
        overview_reduce_code = """function (key, values, rereduce) {
   return sum(values);
}   
"""
        return overview_reduce_code


class TokenSet(object):
    def __init__(self, th=None, tok_config=None):
        print("TokenSet is no longer supported!")

    def create_dict_tokens(
        self,
        iterable={},
        id_prefix="SB",
        id_append="L000000",
        key_name="STARTSB",
        file_upload=None,
    ):
        pass

    def add_attach_to_list(self, attachment, tok_list=None, name=None):
        pass

    @property
    def tokens(self):
        self.update_local_tokens()
        return self.__tokens

    def update_local_tokens(self):
        pass

    def add_keys_to_list(self, key, val, tok_list=None):
        pass
